@@ -13,6 +13,10 @@
 #include <string.h>
 #include <getopt.h>
 #include <locale.h>
+#include <sys/types.h>
+#include <sys/wait.h>
+
+#include <debian-installer.h>
 
 static struct configuration *config = NULL;
 static struct frontend *frontend = NULL;
@@ -28,14 +32,9 @@ static struct option options[] = {
     { 0, 0, 0, 0 },
 };
 
-static void save()
+static int save(void)
 {
-	if (confmodule != NULL)
-		confmodule->update_seen_questions(confmodule, STACK_SEEN_SAVE);
-	if (questions != NULL)
-		questions->methods.save(questions);
-	if (templates != NULL)
-		templates->methods.save(templates);
+	return confmodule->save(confmodule);
 }
 
 static void cleanup()
@@ -50,8 +49,16 @@ static void cleanup()
 		config_delete(confmodule->config);
 }
 
-void sighandler(int sig)
+static void sighandler(int sig)
 {
+	int status = 1;
+	if (sig == SIGCHLD)
+	{
+		while (waitpid(0, &status, WNOHANG) > 0)
+			sigchld_status = status;
+		signal_received = sig;
+		return;
+	}
 	save();
 	/*
 	 * SIGUSR1 used to reconfigure the language. Now it
@@ -60,21 +67,21 @@ void sighandler(int sig)
 	if (sig == SIGUSR1)
 		return;
 	cleanup();
-	exit(1);
+	exit(status);
 }
 
-void help(const char *exename)
+static void help(const char *exename)
 {
     fprintf(stderr, "%s [-ffrontend] [-ppriority] [-oowner] <config script>\n", exename);
     fprintf(stderr, "%s [--frontend=frontend] [--priority=priority] [--owner=owner] <config script>\n", exename);
     exit(-1);
 }
 
-void parsecmdline(struct configuration *config, int argc, char **argv)
+static void parsecmdline(struct configuration *config, int argc, char **argv)
 {
     int c;
 
-    while ((c = getopt_long(argc, argv, "o:p:f:", options, NULL)) > 0)
+    while ((c = getopt_long(argc, argv, "+o:p:f:", options, NULL)) > 0)
     {
         switch (c)
         {
@@ -84,7 +91,7 @@ void parsecmdline(struct configuration *config, int argc, char **argv)
             case 'p':
                 break;
             case 'f':
-		        config->set(config, "frontend::default::driver", optarg);
+                config->set(config, "_cmdline::frontend", optarg);
                 break;
             default:
                 break;
@@ -99,6 +106,13 @@ void parsecmdline(struct configuration *config, int argc, char **argv)
 
 int main(int argc, char **argv)
 {
+	struct sigaction sa;
+
+	sa.sa_handler = &sighandler;
+	sigemptyset(&sa.sa_mask);
+	sa.sa_flags = SA_NOCLDSTOP | SA_RESTART;
+	sigaction(SIGCHLD, &sa, NULL);
+
 	signal(SIGINT, sighandler);
 	signal(SIGTERM, sighandler);
 	signal(SIGUSR1, sighandler);
@@ -147,6 +161,7 @@ int main(int argc, char **argv)
         confmodule->owner = owner;
 	confmodule->run(confmodule, argc - optind + 1, argv + optind - 1);
 	confmodule->communicate(confmodule);
+        confmodule->shutdown(confmodule);
 
 	/* shutting down .... sync the database and shutdown the modules */
 	save();
