@@ -25,7 +25,6 @@
 #include "template.h"
 #include "strutl.h"
 
-#define INFODIR		"/var/lib/dpkg/info"
 #define STATUSFILE	"/var/lib/dpkg/status"
 #define STATUSFIELD	"Status: "
 #define VERSIONFIELD	"Version: "
@@ -34,7 +33,7 @@ struct option g_dpc_args[] = {
 	{ "help", 0, NULL, 'h' },
 	{ "frontend", 1, NULL, 'f' },
 	{ "priority", 1, NULL, 'p' },
-	{ "default-priority", 1, NULL, 'd' },
+	{ "default-priority", 0, NULL, 'd' },
 	{ "all", 0, NULL, 'a' },
 	{ "unseen-only", 0, NULL, 'u' },
 	{ "force", 0, NULL, 'F' },
@@ -93,8 +92,8 @@ static void sighandler(int sig)
  ************************************************************************/
 static void usage(void)
 {
-	printf("dpkg-reconfigure [--frontend <frontend>] [--priority <priority>]\n");
-	printf("dpkg-reconfigure [-f <frontend>] [-p <priority>]\n");
+	printf("dpkg-reconfigure [--frontend <frontend>] [--priority <priority>] package\n");
+	printf("dpkg-reconfigure [-f <frontend>] [-p <priority>] package\n");
 	exit(0);
 }
 
@@ -109,52 +108,13 @@ static void usage(void)
 static bool file_exists(const char *filename, mode_t mode)
 {
 	struct stat buf;
+	if (!filename || !*filename) return false;
 	stat(filename, &buf);
 	if (!S_ISREG(buf.st_mode)) return false;
 	if ((buf.st_mode & mode) == 0) return false;
 	return true;
 }
 
-/************************************************************************
- * Function: loadtemplate
- * Inputs: filename - which file to load templates from
- *         owner - owner for the templates
- * Outputs: none
- * Description: loads all the templates from a file
- * Assumptions: none
- ************************************************************************/
-void loadtemplate(const char *filename, const char *owner)
-{
-	struct template *t;
-	struct question *q;
-
-	t = template_load(filename);
-	while (t)
-	{
-		if (g_templates->methods.set(g_templates, t) != DC_OK)
-			INFO(INFO_ERROR, "Cannot add template %s", t->tag);
-
-		q = g_questions->methods.get(g_questions, t->tag);
-		if (q == NULL)
-		{
-			q = question_new(t->tag);
-			q->template = t;
-			template_ref(t);
-		}
-		else if (q->template != t)
-		{
-			template_deref(q->template);
-			q->template = t;
-			template_ref(t);
-		}
-		question_owner_add(q, owner);
-		if (g_questions->methods.set(g_questions, q) != DC_OK)
-			INFO(INFO_ERROR, "Cannot add question %s", t->tag);
-		question_deref(q);
-		t = t->next;
-	}
-}
-
 /************************************************************************
  * Function: getfield
  * Inputs: package - which package to get the status of
@@ -167,7 +127,8 @@ static char *getfield(const char *package, const char *fieldname)
 {
 	static char field[100] = {0};
 	FILE *fp;
-	char buf[1024], srch[256];
+	char *buf, srch[256];
+	size_t size;
 	int inpkg = 0;
 	int fieldlen;
 
@@ -178,7 +139,8 @@ static char *getfield(const char *package, const char *fieldname)
 
 	if ((fp = fopen(STATUSFILE, "r")) != 0)
 	{
-		while (fgets(buf, sizeof(buf), fp))
+		buf = NULL;
+		while (getline(&buf, &size, fp) > 0)
 		{
 			CHOMP(buf);
 			if (strcmp(buf, srch) == 0)
@@ -199,6 +161,7 @@ static char *getfield(const char *package, const char *fieldname)
 				}
 			}
 		}
+		free(buf);
 		fclose(fp);
 	}
 	return field;
@@ -245,7 +208,7 @@ static int is_confmodule(const char *filename)
  * Description: runs the given confmodule
  * Assumptions: none
  ************************************************************************/
-int runconfmodule(int argc, char **argv)
+static int runconfmodule(int argc, char **argv)
 {
 	struct confmodule *confmodule = NULL;
 	int ret;
@@ -255,11 +218,112 @@ int runconfmodule(int argc, char **argv)
 	confmodule->communicate(confmodule);
 	confmodule->shutdown(confmodule);
 	ret = confmodule->exitcode;
+	g_frontend = confmodule->frontend;
 	confmodule_delete(confmodule);
 
 	return ret;
 }
 
+static char *control_path(const char *pkg, const char *script) {
+	int pipefd[2];
+	int pid;
+
+	if (pipe(pipefd) < 0)
+	{
+		perror("pipe");
+		return NULL;
+	}
+
+	pid = fork();
+	if (pid < 0)
+	{
+		perror("fork");
+		close(pipefd[0]);
+		close(pipefd[1]);
+		return NULL;
+	}
+	else if (pid == 0) /* child */
+	{
+		close(pipefd[0]);
+		dup2(pipefd[1], 1);
+		close(pipefd[1]);
+		execlp("dpkg-query", "dpkg-query", "--control-path",
+		       pkg, script, NULL);
+		/* if we reach here, then execlp failed */
+		perror("execlp");
+		_exit (127);
+	}
+	else /* parent */
+	{
+		FILE *query;
+		char *path = NULL;
+		size_t size;
+
+		close(pipefd[1]);
+		query = fdopen(pipefd[0], "r");
+		getline(&path, &size, query);
+		fclose(query);
+
+		if (path)
+			CHOMP(path);
+
+		return path;
+	}
+}
+
+static int runscript (const char *pkg, const char *script, const char *param) {
+	int ret;
+	char *filename;
+	char *argv[5] = {0};
+	char *version = getfield(pkg, VERSIONFIELD);
+
+	filename = control_path(pkg, script);
+	if (! file_exists(filename,S_IXUSR|S_IXGRP|S_IXOTH))
+	{
+		free(filename);
+		return 0; /* it's ok if the script doesn't exist */
+	}
+	
+	if (strcmp(script, "config") == 0 || is_confmodule(filename))
+	{
+		argv[1] = filename;
+		/* not actually modified, but this ultimately ends up in
+		 * execv(), which wants argv elements to be char *, so
+		 * silence the warning
+		 */
+		argv[2] = (char *) param;
+		argv[3] = version;
+		if ((ret = runconfmodule(4, argv)) != 0)
+		{
+			free(filename);
+			return ret;
+		}
+	}
+	else
+	{
+		/* according to debconf:
+		 * Since a non confmodule might run other programs that
+		 * use debconf, checkpoint the db state and
+		 * reinitialize when the script finishes 
+		 */
+		g_templates->methods.save(g_templates);
+		g_questions->methods.save(g_questions);
+	
+		unsetenv("DEBIAN_HAS_FRONTEND");
+
+		strvacat(filename, sizeof(filename), " ", param, " ", version, NULL);
+		ret = system(filename);
+	
+		setenv("DEBIAN_HAS_FRONTEND", "1", 1);
+	
+		g_templates->methods.load(g_templates);
+		g_questions->methods.load(g_questions);
+	}
+
+	free(filename);
+	return ret;
+}
+
 /************************************************************************
  * Function: reconfigure
  * Inputs: pkgs - pointer to an array of packages
@@ -268,88 +332,37 @@ int runconfmodule(int argc, char **argv)
  * Description: Configure each of the packages passed in
  * Assumptions: none
  ************************************************************************/
-int reconfigure(char **pkgs, int i, int max)
+static int reconfigure(char **pkgs, int i, int max)
 {
 	int ret;
-	char *argv[5] = {0};
-	char filename[1024];
+	char *filename;
 	char *pkg;
 
 	while (i < max)
 	{
 		pkg = pkgs[i++];
 
-		snprintf(filename, sizeof(filename), INFODIR "/%s.config", pkg);
-		if (!file_exists(filename, S_IXUSR|S_IXGRP|S_IXOTH))
-		{
-			INFO(INFO_WARN, "%s is not installed, or does not use debconf", pkg);
-
-			/* Don't die, though this doesn't have a
-			   config script.. we might be doing stuff with
-			   debian-installer, which doesn't use .config
-			   scripts.  Don't uncomment this without talking
-			   to -boot first
-			   -- tfheen, 2002-09-17 */
-
-			/* continue; */
-		}
-		/* startup the confmodule; run the config script and talk to it */
 		g_frontend->methods.set_title(g_frontend, pkg);
 		if (strstr(getfield(pkg, STATUSFIELD), " ok installed") == 0)
 			DIE("%s is not fully installed", pkg);
 
-		snprintf(filename, sizeof(filename), INFODIR "/%s.templates", pkg);
+		filename = control_path(pkg, "templates");
 		if (file_exists(filename, S_IRUSR|S_IRGRP|S_IROTH))
-			loadtemplate(filename, pkg);
+			template_db_loadfile(g_templates, g_questions, filename, pkg, DC_LOADTEMPLATE_NONE);
+		free(filename);
 		
-		snprintf(filename, sizeof(filename), INFODIR "/%s.config", pkg);
-		if (file_exists(filename,S_IXUSR|S_IXGRP|S_IXOTH))
-		{
-			argv[1] = filename;
-			argv[2] = "reconfigure";
-			argv[3] = getfield(pkg, VERSIONFIELD);
-			if ((ret = runconfmodule(4, argv)) != DC_OK)
-				return ret;
-		}
-		else
-		{
-			snprintf(filename, sizeof(filename), INFODIR "/%s.postinst", pkg);
-			if (file_exists(filename,S_IXUSR|S_IXGRP|S_IXOTH))
-			{
-				if (is_confmodule(filename))
-				{
-					argv[1] = filename;
-					argv[2] = "configure";
-					argv[3] = getfield(pkg, VERSIONFIELD);
-					if ((ret = runconfmodule(4, argv)) != 0)
-						return ret;
-				}
-				else
-				{
-					/* according to debconf:
-					 * Since postinst might run other programs that
-					 * use debconf, checkpoint the db state and
-					 * reinitialize when the script finishes 
-					 */
-					g_templates->methods.save(g_templates);
-					g_questions->methods.save(g_questions);
-	
-					unsetenv("DEBIAN_HAS_FRONTEND");
-					setenv("DEBCONF_SHOWOLD", "1", 1);
-
-					strvacat(filename, sizeof(filename), " configure ", getfield(pkg, VERSIONFIELD), NULL);
-	
-					ret = system(filename);
-					if (ret != 0) return DC_NOTOK;
-	
-					setenv("DEBIAN_HAS_FRONTEND", "1", 1);
-					unsetenv("DEBCONF_SHOWOLD");
-	
-					g_templates->methods.load(g_templates);
-					g_questions->methods.load(g_questions);
-				}
-			}
-		}
+		/* Simulation of reinstalling a package, without bothering with
+		   removing the files and putting them back. Just like in a
+		   regular reinstall, run config, and postinst scripts in
+		   sequence, with args. Do not run postrm, because the postrm
+		   can depend on the package's files actually being gone
+		   already. */
+		if ((ret = runscript(pkg, "prerm", "upgrade")) != 0)
+			return ret;
+		if ((ret = runscript(pkg, "config", "reconfigure")) != 0)
+			return ret;
+		if ((ret = runscript(pkg, "postinst", "configure")) != 0)
+			return ret;
 
 		i++;
 	}
@@ -366,6 +379,9 @@ int reconfigure(char **pkgs, int i, int max)
 int main(int argc, char **argv)
 {
 	int opt, ret;
+	int unseen_only=0;
+	int default_priority=0;
+	char *priority_override=NULL;
 
 	signal(SIGINT, sighandler);
 	setlocale (LC_ALL, "");
@@ -375,20 +391,32 @@ int main(int argc, char **argv)
 
 	g_config = config_new();
 
-	while ((opt = getopt_long(argc, argv, "dhf:p:aF", g_dpc_args, NULL)) > 0)
+	while ((opt = getopt_long(argc, argv, "hf:p:duaF", g_dpc_args, NULL)) > 0)
 	{
 		switch (opt)
 		{
 		case 'h': usage(); break;
-		case 'f': setenv("DEBIAN_FRONTEND", optarg, 1); break;
-		case 'p': g_config->set(g_config, "_cmdline::priority", optarg); break;
-		case 'd': break;
-		case 'u': break;
+		case 'f': g_config->set(g_config, "_cmdline::frontend", optarg); break;
+		case 'p': priority_override=strdup(optarg); break;
+		case 'd': default_priority=1; break;
+		case 'u': unseen_only=1; break;
 		case 'a': opt_all = 1; break;
 		case 'F': opt_force = 1; break;
 		}
 	}
-	g_config = config_new();
+	if (optind == argc) {
+		fprintf(stderr, "please specify a package to reconfigure\n");
+		exit(1);
+	}
+	
+	/* Default is to force showing of old questions by default
+	 * for reconfiguring, and show low priority questions. */
+	if (! unseen_only)
+		g_config->set(g_config, "_cmdline::showold", "true");
+	if (priority_override)
+		g_config->set(g_config, "_cmdline::priority", priority_override);
+	else if (! default_priority && (getenv("DEBIAN_PRIORITY") == NULL))
+		g_config->set(g_config, "_cmdline::priority", "low");
 
 	/* parse the configuration info */
 	if (g_config->read(g_config, DEBCONFCONFIG) == 0)
@@ -397,14 +425,16 @@ int main(int argc, char **argv)
 	/* initialize database and frontend modules */
 	if ((g_templates = template_db_new(g_config, NULL)) == 0)
 		DIE("Cannot initialize DebConf templates database");
+	if (g_templates->methods.load(g_templates) == 0)
+		DIE("Cannot initialize DebConf templates database");
 	if ((g_questions = question_db_new(g_config, g_templates, NULL)) == 0)
 		DIE("Cannot initialize DebConf config database");
+	if (g_questions->methods.load(g_questions) == 0)
+		DIE("Cannot initialize DebConf config database");
+
 	if ((g_frontend = frontend_new(g_config, g_templates, g_questions)) == 0)
 		DIE("Cannot initialize DebConf frontend");
 
-	g_templates->methods.load(g_templates);
-	g_questions->methods.load(g_questions);
-
 	setenv("DEBIAN_HAS_FRONTEND", "1", 1);
 	ret = reconfigure(argv, optind, argc);
 	/* shutting down .... sync the database and shutdown the modules */
