@@ -3,8 +3,6 @@
  * cdebconf - An implementation of the Debian Configuration Management
  *            System
  *
- * $Id$
- *
  * cdebconf is (c) 2000-2007 Randolph Chung and others under the following
  * license.
  *
@@ -104,7 +102,7 @@ static gboolean handle_exposed_banner(GtkWidget * widget,
     char * text;
 
     if (NULL != fe->info) {
-        text = q_get_description(fe->info);
+        text = q_get_description(fe, fe->info);
         message = g_strdup_printf(
             "<b><span foreground=\"#ffffff\">%s</span></b>", text);
         layout = gtk_widget_create_pango_layout(
@@ -303,8 +301,6 @@ gboolean cdebconf_gtk_create_main_window(struct frontend * fe)
     gtk_window_set_decorated(GTK_WINDOW(window), TRUE /* resizable */);
     create_main_widgets(fe, window);
 
-    /* XXX: we might want to unconnect this signal in case of normal
-     * shutdown... */
     g_signal_connect_swapped(window, "destroy",
                              G_CALLBACK(handle_closed_main_window), fe);
 
@@ -367,6 +363,8 @@ void cdebconf_gtk_destroy_main_window(struct frontend * fe)
     GtkWidget * window = fe_data->window;
 
     if (NULL != window) {
+        g_signal_handlers_disconnect_by_func(
+            window, G_CALLBACK(handle_closed_main_window), fe);
         fe_data->window = NULL;
         g_object_unref(G_OBJECT(window));
         gtk_widget_destroy(window);
@@ -454,6 +452,7 @@ gboolean cdebconf_gtk_run_message_dialog(struct frontend * fe,
     GtkWidget * dialog;
     GtkWidget * vbox;
     GtkWidget * frame;
+    GtkWidget * label;
 
     /* XXX: check NULL! */
     dialog = gtk_window_new(GTK_WINDOW_TOPLEVEL);
@@ -465,13 +464,17 @@ gboolean cdebconf_gtk_run_message_dialog(struct frontend * fe,
     gtk_window_set_decorated(GTK_WINDOW(dialog), FALSE /* no decoration */);
     gtk_container_set_border_width(GTK_CONTAINER(dialog), 0 /* no border */);
 
+    /* Create a wrapped label */
+    label = gtk_label_new(message);
+    gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
+
     /* XXX: check NULL! */
     vbox = gtk_vbox_new(FALSE /* don't make children equal */,
                         DEFAULT_PADDING);
     gtk_box_pack_start(GTK_BOX(vbox), create_dialog_title_label(title),
                        FALSE /* don't expand */, FALSE /* don't fill */,
                        0 /* no padding */);
-    gtk_box_pack_start(GTK_BOX(vbox), gtk_label_new(message),
+    gtk_box_pack_start(GTK_BOX(vbox), label,
                        FALSE /* don't expand */, FALSE /* don't fill */,
                        DEFAULT_PADDING);
     gtk_box_pack_start(GTK_BOX(vbox), gtk_hseparator_new(),
@@ -593,5 +596,23 @@ void cdebconf_gtk_empty_target_box(struct frontend * fe)
                          (GtkCallback) gtk_widget_destroy, NULL /* no data */);
 }
 
+/** Returns the width (in pango units) of the given text.
+ *
+ * @param widget widget where the text will be renderered
+ * @param text text to be rendered
+ * @return width of the text in pango units
+ */
+gint cdebconf_gtk_get_text_width(GtkWidget * widget, gchar * text)
+{
+    PangoLayout * layout;
+    gint width;
+
+    layout = gtk_widget_create_pango_layout(widget, text);
+    pango_layout_get_size(layout, &width, NULL /* no height */);
+    g_object_unref(layout);
+
+    return width;
+}
+
 /* vim: et sw=4 si
  */
