@@ -7,8 +7,6 @@
  *
  * Description: Newt UI for cdebconf
  *
- * $Id: newt.c 49677 2007-10-02 17:12:21Z lunar $
- *
  * cdebconf is (c) 2000-2001 Randolph Chung and others under the following
  * license.
  *
@@ -71,48 +69,62 @@ struct newt_data {
                   scale_cancel,
                   perc_label;
     int           scale_textbox_height;
-    char *        scale_info;
+    struct question *scale_info;
 };
 
 struct newtColors newtAltColorPalette = {
-	"white", "blue", 			/* root fg, bg */
-	/*"black", "lightgray",*/			/* border fg, bg */
+	/*"white", "blue",*/ 			/* root fg, bg */
+	"white", "black", 			/* root fg, bg */
+	/*"black", "lightgray",*/		/* border fg, bg */
 	"white", "black",			/* border fg, bg */
-	/*"black", "lightgray",*/			/* window fg, bg */
+	/*"black", "lightgray",*/		/* window fg, bg */
 	"white", "black",			/* window fg, bg */
 	/*"white", "black",*/			/* shadow fg, bg */
-	"white", "blue",			/* shadow fg, bg */
+	"white", "black",			/* shadow fg, bg */
 	/*"red", "lightgray",*/			/* title fg, bg */
 	"yellow", "black",			/* title fg, bg */
-	"brightred", "gray",			/* button fg, bg */
+	/*"brightred", "gray",*/		/* button fg, bg */
+	"black", "lightgray",			/* button fg, bg */
 	"gray", "brightred",			/* active button fg, bg */
 	"white", "gray",			/* checkbox fg, bg */
-	"yellow", "brown",			/* active checkbox fg, bg */
-	"yellow", "blue",			/* entry box fg, bg */
-	/*"blue", "lightgray",*/			/* label fg, bg */
+	/*"yellow", "brown",*/			/* active checkbox fg, bg */
+	"black", "lightgray",			/* active checkbox fg, bg */
+	/*"yellow", "blue",*/			/* entry box fg, bg */
+	"white", "black",			/* entry box fg, bg */
+	/*"blue", "lightgray",*/		/* label fg, bg */
 	"brightred", "black",			/* label fg, bg */
-	/*"black", "lightgray",*/			/* listbox fg, bg */
-	"lightgray", "black",			/* listbox fg, bg */
+	/*"black", "lightgray",*/		/* listbox fg, bg */
+	"white", "black",			/* listbox fg, bg */
 	"yellow", "blue",			/* active listbox fg, bg */
-	/*"black", "lightgray",*/			/* textbox fg, bg */
+	/*"black", "lightgray",*/		/* textbox fg, bg */
 	"white", "black",			/* textbox fg, bg */
 	"lightgray", "black",			/* active textbox fg, bg */
 	/*"white", "blue",*/			/* help line */
 	"white", "black",			/* help line */
 	"yellow", "blue",			/* root text */
-	"blue",					/* scale full */
-	"red",					/* scale empty */
+	/*"blue",*/				/* scale full */
+	"black",				/* scale full */
+	/*"red",*/				/* scale empty */
+	"blue",					/* scale empty */
 	"blue", "lightgray",			/* disabled entry fg, bg */
-	/*"black", "lightgray",*/			/* compact button fg, bg */
-	"lightgray", "black",			/* compact button fg, bg */
-	"yellow", "gray",			/* active & sel listbox */
+	/*"black", "lightgray",*/		/* compact button fg, bg */
+	"white", "black",			/* compact button fg, bg */
+	"black", "lightgray",			/* active & sel listbox */
 	"black", "brown"			/* selected listbox */
 };
 
 typedef int (newt_handler)(struct frontend *obj, struct question *q);
 
+struct newt_help_callback_data {
+    struct frontend *obj;
+    const char *tag;
+};
+
 static void newt_progress_stop(struct frontend *obj);
 
+/* Result must be freed by the caller */
+static char *get_full_description(struct frontend *obj, struct question *q);
+
 #include "cdebconf_newt.h"
 
 /*  Padding of title width, allows for leading "[!!] " before title
@@ -120,6 +132,14 @@ static void newt_progress_stop(struct frontend *obj);
  */
 #define TITLE_PADDING 9
 
+/*  At least one blue line on top
+ *  One line for the title / top frame
+ *  One line for the bottom frame
+ *  One line for shadow on the bottom
+ *  One line with a menu
+ */
+#define MIN_DECORATION_HEIGHT 5
+
 /* gettext would be much nicer :-( */
 static const char *
 continue_text(struct frontend *obj)
@@ -154,7 +174,47 @@ cancel_text(struct frontend *obj)
 static const char *
 help_text(struct frontend *obj)
 {
-    return question_get_text(obj, "debconf/help-line", "<Tab> moves between items; <Space> selects; <Enter> activates buttons");
+    return question_get_text(obj, "debconf/help-line", "<Tab> moves; <Space> selects; <Enter> activates buttons");
+}
+
+static const char *
+help_text_f1(struct frontend *obj)
+{
+    return question_get_text(obj, "debconf/help-line-f1", "<F1> for help; <Tab> moves; <Space> selects; <Enter> activates buttons");
+}
+
+void
+cdebconf_newt_setup(void)
+{
+    SLang_init_tty(0, 1, 0); /* disable flow control */
+    newtInit();
+    newtSetColors(newtAltColorPalette);
+    newtCls();
+}
+
+/* cdebconf-newt-terminal needs this in order to be able to restore the
+ * display properly after tearing down the terminal.
+ */
+struct question *
+cdebconf_newt_get_progress_info(struct frontend *obj)
+{
+    struct newt_data *data = (struct newt_data *)obj->data;
+    if (data->scale_info) {
+        question_ref(data->scale_info);
+        return data->scale_info;
+    } else
+        return NULL;
+}
+
+/* This is global due to cdebconf_newt_create_form's interface. It might be
+ * worth fixing this at the next convenient API break.
+ */
+static struct newt_help_callback_data *help_cb_data;
+
+newtComponent
+cdebconf_newt_create_form(newtComponent scrollbar)
+{
+    return newtForm(scrollbar, help_cb_data, 0);
 }
 
 void
@@ -245,17 +305,29 @@ cdebconf_newt_get_text_width(const char *text)
 }
 
 static int
-min_window_height(struct question *q, int win_width)
+min_window_height(struct frontend *obj, struct question *q, int win_width)
 {
+    // start with a blank or description (note and error)
+    // End with <Continue>/boolean buttons + blank
     int height = 3;
     char *type = q->template->type;
-    char *q_ext_text;
+    char *q_text;
 
-    q_ext_text = q_get_extended_description(q);
-    if (q_ext_text != NULL)
-        height = cdebconf_newt_get_text_height(q_ext_text, win_width) + 1;
-    if (strcmp(type, "multiselect") == 0 || strcmp(type, "select") == 0)
-        height += 4; // at least three lines for choices + blank line
+    if (strcmp(q->template->type, "note") == 0 || strcmp(q->template->type, "error") == 0)
+        q_text = q_get_extended_description(obj, q);
+    else
+        q_text = get_full_description(obj, q);
+    if (q_text != NULL) {
+        height += cdebconf_newt_get_text_height(q_text, win_width) + 1;
+        free (q_text);
+    }
+    if (strcmp(type, "multiselect") == 0)
+        height += 4; // x lines for choices + blank line
+    else if (strcmp(type, "select") == 0) {
+        height += 2; // as multiselect, but without a button and blank line
+        if (obj->methods.can_go_back(obj, q))
+            height += 2;
+    }
     else if (strcmp(type, "string") == 0 || strcmp(type, "password") == 0)
         height += 2; // input line + blank line
     // the others don't need more space
@@ -263,22 +335,22 @@ min_window_height(struct question *q, int win_width)
 }
 
 static int
-need_separate_window(struct question *q)
+need_separate_window(struct frontend *obj, struct question *q)
 {
     int width = 80, height = 24;
     int x;
 
     newtGetScreenSize(&width, &height);
-    x = min_window_height(q, width-7);
-    return (x >= height-5);
+    x = min_window_height(obj, q, width-7);
+    return (x > height-MIN_DECORATION_HEIGHT);
 }
 
 static char *
-get_full_description(struct question *q)
+get_full_description(struct frontend *obj, struct question *q)
 {
     char *res = NULL;
-    char *descr = q_get_description(q);
-    char *ext_descr = q_get_extended_description(q);
+    char *descr = q_get_description(obj, q);
+    char *ext_descr = q_get_extended_description(obj, q);
 
     assert(descr);
     assert(ext_descr);
@@ -295,7 +367,7 @@ get_full_description(struct question *q)
 }
 
 static int
-show_separate_window(struct frontend *obj, struct question *q)
+show_separate_window(struct frontend *obj, struct question *q, int is_help)
 {
     newtComponent form, textbox, bOk, bCancel, cRet;
     int width = 80, height = 24, t_height, t_width, win_width, win_height;
@@ -303,6 +375,7 @@ show_separate_window(struct frontend *obj, struct question *q)
     // buttons
     int extra = 3;
     int format_note = 0;
+    int t_voffset = 0;
     char *full_description;
     int ret;
 #ifdef HAVE_LIBTEXTWRAP
@@ -312,8 +385,8 @@ show_separate_window(struct frontend *obj, struct question *q)
 #else
     int flags = NEWT_FLAG_WRAP;
 #endif
-    char *descr = q_get_description(q);
-    char *ext_descr = q_get_extended_description(q);
+    char *descr = q_get_description(obj, q);
+    char *ext_descr = q_get_extended_description(obj, q);
 
     assert(descr);
     assert(ext_descr);
@@ -321,11 +394,12 @@ show_separate_window(struct frontend *obj, struct question *q)
     if (strcmp(q->template->type, "note") == 0 || strcmp(q->template->type, "error") == 0)
     {
         format_note = 1;
-        extra++;
+        t_voffset = 1;
+        extra += 2;
         full_description = strdup(ext_descr);
     }
     else
-        full_description = get_full_description(q);
+        full_description = get_full_description(obj, q);
 
     newtGetScreenSize(&width, &height);
     win_width = width-7;
@@ -352,7 +426,7 @@ show_separate_window(struct frontend *obj, struct question *q)
     if (t_width_descr > t_width)
         t_width = t_width_descr;
     t_width_buttons = 2*BUTTON_PADDING + cdebconf_newt_get_text_width(continue_text(obj)) + 2;
-    if (obj->methods.can_go_back(obj, q))
+    if (!is_help && obj->methods.can_go_back(obj, q))
         //  Add an interspace
         t_width_buttons += cdebconf_newt_get_text_width(goback_text(obj)) + 3;
     if (t_width_buttons > t_width)
@@ -365,12 +439,12 @@ show_separate_window(struct frontend *obj, struct question *q)
     cdebconf_newt_create_window(win_width, win_height, obj->title, q->priority);
     form = cdebconf_newt_create_form(NULL);
     if (format_note)
-        newtFormAddComponent(form, newtLabel((win_width - strwidth(descr))/2, 0, descr));
-    textbox = newtTextbox(TEXT_PADDING, 1, t_width, t_height, flags);
+        newtFormAddComponent(form, newtLabel((win_width - strwidth(descr))/2, 1, descr));
+    textbox = newtTextbox(TEXT_PADDING, 1 + t_voffset, t_width, t_height, flags);
     assert(textbox);
     newtTextboxSetText(textbox, full_description);
     free(full_description);
-    if (obj->methods.can_go_back(obj, q)) {
+    if (!is_help && obj->methods.can_go_back(obj, q)) {
         bOk     = newtCompactButton(win_width - TEXT_PADDING - BUTTON_PADDING - strwidth(continue_text(obj)) - 3, win_height-2, continue_text(obj));
         bCancel = newtCompactButton(TEXT_PADDING + BUTTON_PADDING - 1,  win_height-2, goback_text(obj));
         newtFormAddComponents(form, bCancel, textbox, bOk, NULL);
@@ -410,7 +484,7 @@ generic_handler_string(struct frontend *obj, struct question *q, int eflags)
 #endif
     char *defval;
     const char *result;
-    char *full_description = get_full_description(q);
+    char *full_description = get_full_description(obj, q);
 
     eflags |= NEWT_ENTRY_SCROLL | NEWT_FLAG_RETURNEXIT;
     newtGetScreenSize(&width, &height);
@@ -496,8 +570,8 @@ show_multiselect_window(struct frontend *obj, struct question *q, int show_ext_d
     size_t res;
     wchar_t c;
     int *tindex = NULL;
-    const char *indices = q_get_indices(q);
-    char *full_description = get_full_description(q);
+    const char *indices = q_get_indices(obj, q);
+    char *full_description = get_full_description(obj, q);
 #ifdef HAVE_LIBTEXTWRAP
     textwrap_t tw;
     char *wrappedtext;
@@ -509,15 +583,19 @@ show_multiselect_window(struct frontend *obj, struct question *q, int show_ext_d
     newtGetScreenSize(&width, &height);
     win_width = width-7;
     strtruncate(obj->title, win_width-9);
-    count = strgetargc(q_get_choices_vals(q));
+    count = strgetargc(q_get_choices_vals(obj, q));
     if (count <= 0)
         return DC_NOTOK;
     choices = malloc(sizeof(char *) * count);
     choices_trans = malloc(sizeof(char *) * count);
     tindex = malloc(sizeof(int) * count);
-    if (strchoicesplitsort(q_get_choices_vals(q), q_get_choices(q), indices, choices, choices_trans, tindex, count) != count)
+    if (strchoicesplitsort(q_get_choices_vals(obj, q), q_get_choices(obj, q), indices, choices, choices_trans, tindex, count) != count)
         return DC_NOTOK;
 
+    if (obj->methods.can_align(obj, q)) {
+        stralign(choices_trans, count);
+    }
+
     defvals = malloc(sizeof(char *) * count);
     defcount = strchoicesplit(question_getvalue(q, ""), defvals, count);
     answer = malloc(sizeof(char) * count);
@@ -644,12 +722,14 @@ show_select_window(struct frontend *obj, struct question *q, int show_ext_desc)
     int listflags = NEWT_FLAG_RETURNEXIT;
     int width = 80, height = 24;
     int win_width, win_height = -1, t_height, t_width, sel_height, sel_width;
+    int select_list_top;
+    int b_height;
     int t_width_title, t_width_buttons;
     char **choices, **choices_trans, *defval;
     int count = 0, i, ret, defchoice = -1;
     int *tindex = NULL;
-    const char *indices = q_get_indices(q);
-    char *full_description = get_full_description(q);
+    const char *indices = q_get_indices(obj, q);
+    char *full_description = get_full_description(obj, q);
     char *p;
     size_t res;
     int k;
@@ -665,15 +745,19 @@ show_select_window(struct frontend *obj, struct question *q, int show_ext_desc)
     newtGetScreenSize(&width, &height);
     win_width = width-7;
     strtruncate(obj->title, win_width-9);
-    count = strgetargc(q_get_choices_vals(q));
+    count = strgetargc(q_get_choices_vals(obj, q));
     if (count <= 0)
         return DC_NOTOK;
     choices = malloc(sizeof(char *) * count);
     choices_trans = malloc(sizeof(char *) * count);
     tindex = malloc(sizeof(int) * count);
-    if (strchoicesplitsort(q_get_choices_vals(q), q_get_choices(q), indices, choices, choices_trans, tindex, count) != count)
+    if (strchoicesplitsort(q_get_choices_vals(obj, q), q_get_choices(obj, q), indices, choices, choices_trans, tindex, count) != count)
         return DC_NOTOK;
 
+    if (obj->methods.can_align(obj, q)) {
+        stralign(choices_trans, count);
+    }
+
     sel_height = count;
     form = cdebconf_newt_create_form(NULL);
 #ifdef HAVE_LIBTEXTWRAP
@@ -719,18 +803,28 @@ show_select_window(struct frontend *obj, struct question *q, int show_ext_desc)
         t_height = newtTextboxGetNumLines(textbox);
         newtTextboxSetHeight(textbox, t_height);
         newtFormAddComponent(form, textbox);
-    } else
+        select_list_top = 1+t_height+1;
+    } else {
         t_height = 0;
+        select_list_top = 1; // No description. Only insert a blank line.
+    }
+    if (obj->methods.can_go_back(obj, q))
+        b_height = 2;
+    else
+        b_height = 0;
     free(full_description);
-    win_height = t_height + 5 + sel_height;
-    if (win_height > height-5) {
-        win_height = height-5;
-        sel_height = win_height - t_height - 5;
+    win_height  = t_height + sel_height + b_height;
+    //    3 == First blank line + blanks before and after select
+    // or 3 == Blanks before and after select + blank after <Go Back>
+    win_height += 3;
+    if (win_height > height-MIN_DECORATION_HEIGHT) {
+        win_height = height-MIN_DECORATION_HEIGHT;
+        sel_height = win_height - t_height - 3 - b_height;
     }
     if (count > sel_height)
         listflags |= NEWT_FLAG_SCROLL;
     cdebconf_newt_create_window(win_width, win_height, obj->title, q->priority);
-    listbox = newtListbox((win_width-sel_width-3)/2, 1+t_height+1, sel_height, listflags);
+    listbox = newtListbox((win_width-sel_width-3)/2, select_list_top, sel_height, listflags);
     defval = (char *)question_getvalue(q, "");
     for (i = 0; i < count; i++) {
         newtListboxAppendEntry(listbox, choices_trans[i], choices[tindex[i]]);
@@ -788,7 +882,7 @@ newt_handler_boolean(struct frontend *obj, struct question *q)
 #else
     int flags = NEWT_FLAG_WRAP;
 #endif
-    char *full_description = get_full_description(q);
+    char *full_description = get_full_description(obj, q);
 
     newtGetScreenSize(&width, &height);
     win_width = width-7;
@@ -878,10 +972,10 @@ newt_handler_multiselect(struct frontend *obj, struct question *q)
 {
     int separate_window, ret;
 
-    separate_window = need_separate_window(q);
+    separate_window = need_separate_window(obj, q);
     while (1) {
         if (separate_window) {
-            ret = show_separate_window(obj, q);
+            ret = show_separate_window(obj, q, 0);
             if (ret != DC_OK)
                 return ret;
             ret = show_multiselect_window(obj, q, 0);
@@ -906,10 +1000,10 @@ newt_handler_select(struct frontend *obj, struct question *q)
 {
     int separate_window, ret;
 
-    separate_window = need_separate_window(q);
+    separate_window = need_separate_window(obj, q);
     while (1) {
         if (separate_window) {
-            ret = show_separate_window(obj, q);
+            ret = show_separate_window(obj, q, 0);
             if (ret != DC_OK)
                 return ret;
             ret = show_select_window(obj, q, 0);
@@ -959,7 +1053,7 @@ static int newt_handler_password(struct frontend *obj, struct question *q)
 static int newt_handler_note(struct frontend *obj, struct question *q)
 {
     // Sort of hack-ish, yes, but it works :-)
-    return show_separate_window(obj, q);
+    return show_separate_window(obj, q, 0);
 }
 
 /*
@@ -1003,6 +1097,23 @@ newt_handler_error(struct frontend *obj, struct question *q)
     return ret;
 }
 
+static void
+newt_help_callback(newtComponent co, void *d)
+{
+    struct newt_help_callback_data *data = d;
+    struct frontend *obj = data->obj;
+    struct question *q = obj->qdb->methods.get(obj->qdb, data->tag);
+
+    /* Don't allow recursive calls. */
+    newtSetHelpCallback(NULL);
+
+    show_separate_window(obj, q, 1);
+
+    newtSetHelpCallback(newt_help_callback);
+
+    question_deref(q);
+}
+
 /* ----------------------------------------------------------------------- */
 struct question_handlers {
 	const char *type;
@@ -1033,7 +1144,7 @@ struct question_handlers {
 static int
 newt_initialize(struct frontend *obj, struct configuration *conf)
 {
-    int i, width = 80, height = 24;
+    int width = 80, height = 24;
     const char *palette;
 
     obj->interactive = 1;
@@ -1045,9 +1156,6 @@ newt_initialize(struct frontend *obj, struct configuration *conf)
         newtAltColorPalette = newtDefaultColorPalette;
     newtSetColors(newtAltColorPalette);
     newtGetScreenSize(&width, &height);
-    // Fill the screen so people can shift-pgup properly
-    for (i = 0; i < height; i++)
-        putchar('\n');
     newtFinished();
     return DC_OK;
 }
@@ -1087,44 +1195,55 @@ newt_go(struct frontend *obj)
             }
 
             if (plugin || strcmp(q->template->type, question_handlers[i].type) == 0) {
+                char *help = q_get_help(obj, q);
+                struct newt_help_callback_data my_help_cb_data;
                 if (!cleared && !data->scale_form) {
                     cleared = 1;
-                    SLang_init_tty(0, 1, 0); /* disable flow control */
-                    newtInit();
-                    newtSetColors(newtAltColorPalette);
-                    newtCls();
+                    cdebconf_newt_setup();
+                }
+                if (*help) {
+                    my_help_cb_data.obj = obj;
+                    my_help_cb_data.tag = help;
+                    help_cb_data = &my_help_cb_data;
+                    newtSetHelpCallback(newt_help_callback);
                 }
                 if (obj->info != NULL) {
-                    char *text = q_get_description(obj->info);
+                    char *text = q_get_description(obj, obj->info);
                     if (text)
                         newtDrawRootText(0, 0, text);
                     free(text);
                 }
-                newtPushHelpLine(help_text(obj));
+                if (*help)
+                    newtPushHelpLine(help_text_f1(obj));
+                else
+                    newtPushHelpLine(help_text(obj));
                 ret = handler(obj, q);
                 newtPopHelpLine();
                 if (ret == DC_OK)
                     obj->qdb->methods.set(obj->qdb, q);
-                else if (ret == DC_GOBACK && q->prev != NULL)
-                    q = q->prev;
-                else {
-                    if (cleared && !data->scale_form)
-                        newtFinished();
-                    if (plugin)
-                        plugin_delete(plugin);
-                    return ret;
+                if (*help) {
+                    newtSetHelpCallback(NULL);
+                    help_cb_data = NULL;
                 }
+                free(help);
                 if (plugin)
                     plugin_delete(plugin);
                 break;
             }
         }
+        if (ret == DC_NOTOK)
+            break;
         if (ret == DC_OK)
             q = q->next;
+        else if (ret == DC_GOBACK) {
+            do {
+                q = q->prev;
+            } while (q != NULL && 0 == strcmp("error", q->template->type));
+	}
     }
     if (cleared && !data->scale_form)
         newtFinished();
-    return DC_OK;
+    return ret;
 }
 
 static bool
@@ -1139,13 +1258,36 @@ newt_can_cancel_progress(struct frontend *obj)
     return (obj->capability & DCF_CAPB_PROGRESSCANCEL);
 }
 
+static bool
+newt_can_align(struct frontend *obj, struct question *q)
+{
+    return (obj->capability & DCF_CAPB_ALIGN);
+}
+
+static const char *
+newt_lookup_directive(struct frontend *obj, const char *directive)
+{
+    if (obj->methods.can_align(obj, obj->questions)) {
+        if (strcmp("TAB", directive) == 0) {
+            return STRALIGN_TAB;
+        } else if (strcmp("ALIGN=CENTER", directive) == 0) {
+            return STRALIGN_ALIGN_CENTER;
+        } else if (strcmp("ALIGN=RIGHT", directive) == 0) {
+            return STRALIGN_ALIGN_RIGHT;
+        }
+    }
+    /* Remove unhandled directives */
+    return "";
+}
+
 static void
-newt_make_progress_bar(struct frontend *obj, const char *info)
+newt_make_progress_bar(struct frontend *obj, struct question *info)
 {
     struct newt_data *data = (struct newt_data *)obj->data;
     int width = 80, height = 24, win_width, win_height, text_height;
     int extra = 0;
     bool can_cancel_progress;
+    char *info_desc = NULL;
     bool changed = false;
 #ifdef HAVE_LIBTEXTWRAP
     textwrap_t tw;
@@ -1162,10 +1304,14 @@ newt_make_progress_bar(struct frontend *obj, const char *info)
     win_width = width-7;
     if (!info)
         info = data->scale_info;
-    else
-        data->scale_info = strdup(info);
+    else {
+        data->scale_info = info;
+        question_ref(data->scale_info);
+    }
     if (info)
-        text_height = cdebconf_newt_get_text_height(info, win_width);
+        info_desc = q_get_raw_description(info);
+    if (info_desc)
+        text_height = cdebconf_newt_get_text_height(info_desc, win_width);
     else
         text_height = 0;
     /* Minimal height set to 2 to prevent box flashing */
@@ -1184,11 +1330,14 @@ newt_make_progress_bar(struct frontend *obj, const char *info)
         }
     }
     if (!data->scale_form || changed) {
+        char *title_desc;
         if (text_height + 3 + extra <= height - 5)
             win_height = text_height + 3 + extra;
         else
             win_height = height - 5;
-        newtCenteredWindow(win_width, win_height, obj->progress_title);
+        title_desc = q_get_raw_description(obj->progress_title);
+        newtCenteredWindow(win_width, win_height, title_desc);
+        free(title_desc);
         data->scale_bar = newtScale(TEXT_PADDING, 1, win_width-2*TEXT_PADDING, obj->progress_max - obj->progress_min);
         data->scale_textbox = newtTextbox(TEXT_PADDING, 3, win_width-2*TEXT_PADDING, text_height, flags);
         data->scale_textbox_height = text_height;
@@ -1202,21 +1351,23 @@ newt_make_progress_bar(struct frontend *obj, const char *info)
         newtFormSetTimer(data->scale_form, 1);
     }
     newtScaleSet(data->scale_bar, obj->progress_cur - obj->progress_min);
-    if (info) {
+    if (info_desc) {
 #ifdef HAVE_LIBTEXTWRAP
         textwrap_init(&tw);
         textwrap_columns(&tw, win_width - 2 - 2*TEXT_PADDING);
-        wrappedtext = textwrap(&tw, info);
+        wrappedtext = textwrap(&tw, info_desc);
         newtTextboxSetText(data->scale_textbox, wrappedtext);
         free(wrappedtext);
 #else
-        newtTextboxSetText(data->scale_textbox, info);
+        newtTextboxSetText(data->scale_textbox, info_desc);
 #endif
     }
+    free(info_desc);
 }
 
 static void
-newt_progress_start(struct frontend *obj, int min, int max, const char *title)
+newt_progress_start(struct frontend *obj, int min, int max,
+                    struct question *title)
 {
     struct newt_data *data = (struct newt_data *)obj->data;
 
@@ -1225,18 +1376,16 @@ newt_progress_start(struct frontend *obj, int min, int max, const char *title)
          * use one anyway, tear down the old progress bar first.
          */
         newt_progress_stop(obj);
-    DELETE(obj->progress_title);
-    obj->progress_title = strdup(title);
+    question_deref(obj->progress_title);
+    obj->progress_title = title;
+    question_ref(obj->progress_title);
     obj->progress_min = min;
     obj->progress_max = max;
     obj->progress_cur = min;
     data->scale_info = NULL;
-    SLang_init_tty(0, 1, 0); /* disable flow control */
-    newtInit();
-    newtSetColors(newtAltColorPalette);
-    newtCls();
+    cdebconf_newt_setup();
     if (obj->info != NULL) {
-        char *text = q_get_description(obj->info);
+        char *text = q_get_description(obj, obj->info);
         if (text)
             newtDrawRootText(0, 0, text);
         free(text);
@@ -1283,7 +1432,7 @@ newt_progress_set(struct frontend *obj, int val)
 }
 
 static int
-newt_progress_info(struct frontend *obj, const char *info)
+newt_progress_info(struct frontend *obj, struct question *info)
 {
     struct newt_data *data = (struct newt_data *)obj->data;
     struct newtExitStruct es;
@@ -1316,7 +1465,7 @@ newt_progress_stop(struct frontend *obj)
         newtPopWindow();
         newtFinished();
         data->scale_form = data->scale_bar = data->perc_label = data->scale_textbox = data->scale_cancel = NULL;
-        free(data->scale_info);
+        question_deref(data->scale_info);
         data->scale_info = NULL;
     }
 }
@@ -1333,9 +1482,11 @@ struct frontend_module debconf_frontend_module =
 {
 initialize: newt_initialize,
 shutdown: newt_shutdown,
+lookup_directive: newt_lookup_directive,
 go: newt_go,
 can_go_back: newt_can_go_back,
 can_cancel_progress: newt_can_cancel_progress,
+can_align: newt_can_align,
 progress_start: newt_progress_start,
 progress_set:   newt_progress_set,
 progress_info:  newt_progress_info,
